import os
import pkg_resources

from nanite.rate import io as nio
from PyQt5 import uic, QtWidgets


# load QWidget from ui file
ui_path = pkg_resources.resource_filename("pyjibe.fd",
                                          "user_rating_design.ui")
UiUserRatingBase = uic.loadUiType(ui_path)[0]


class Rater(QtWidgets.QWidget, UiUserRatingBase):
    def __init__(self, fdui, path):
        QtWidgets.QWidget.__init__(self, None)
        UiUserRatingBase.__init__(self)
        self.setupUi(self)
        self.fdui = fdui
        self.path = path

        self.initial_stuff()
        self.setup_signals()
        self.on_change_index()  # load first dataset if applicable

    def initial_stuff(self):
        # set container path
        self.container_path.setText(str(self.path))
        # set user name
        for name in ('LOGNAME', 'USER', 'LNAME', 'USERNAME'):
            if name in os.environ:
                user = os.environ[name]
                break
        else:
            user = "unknown"
        self.user_name.setText(user)
        # set correct initial index
        item = self.fdui.list_curves.currentItem()
        idx = self.fdui.list_curves.indexOfTopLevelItem(item)
        self.curve_index.setValue(idx+1)
        self.curve_index.setMaximum(len(self.fdui.data_set))

    def setup_signals(self, enable=True):
        cn = [
            # Change of index updates apc
            [self.curve_index.valueChanged, self.on_change_index],
            [self.sp_rating.valueChanged, self.on_change_values],
            [self.text_comment.textChanged, self.on_change_values],
        ]

        for signal, slot in cn:
            if enable:
                signal.connect(slot)
            else:
                signal.disconnect(slot)

    def on_change_index(self):
        index = self.curve_index.value()-1
        self.setup_signals(False)
        fdist = self.fdui.data_set[index]
        _, rating, comment = nio.hdf5_rated(self.path, fdist)
        self.sp_rating.setValue(rating)
        self.text_comment.setPlainText(comment)
        self.setup_signals(True)
        it = self.fdui.list_curves.topLevelItem(index)
        self.fdui.list_curves.setCurrentItem(it)
        self.sp_rating.selectAll()
        self.sp_rating.setFocus()

    def on_change_values(self):
        index = self.curve_index.value()-1
        fdist = self.fdui.data_set[index]
        nio.save_hdf5(h5path=self.path,
                      indent=fdist,
                      user_rate=self.sp_rating.value(),
                      user_name=self.user_name.text(),
                      user_comment=self.text_comment.toPlainText())



"""Settings file management"""
import pathlib

import appdirs

#: default settings file name
NAME = "pyjibe.cfg"

#: default configuration parameters
DEFAULTS = {}


class SettingsFile(object):
    """Manages a settings file in the user's config dir"""

    def __init__(self, name=NAME, defaults=DEFAULTS, directory=None):
        """Initialize settings file (create if it does not exist)"""
        if directory is None:
            directory = appdirs.user_config_dir()
        fname = pathlib.Path(directory) / name
        # create file if not existent
        if not fname.exists():
            # Create the file
            fname.open("w").close()

        self.cfgfile = fname
        self.defaults = defaults
        self.working_directories = {}

    def load(self):
        """Loads the settings file returning a dictionary"""
        with self.cfgfile.open() as fop:
            fc = fop.readlines()
        cdict = {}
        for line in fc:
            line = line.strip()
            var, val = line.split("=", 1)
            cdict[var.lower().strip()] = val.strip()
        return cdict

    def get_bool(self, key):
        """Returns boolean configuration key"""
        key = key.lower()
        cdict = self.load()
        if key in cdict:
            val = cdict[key]
            if val.lower() not in ["true", "false"]:
                raise ValueError("Config key '{}' not boolean!".format(key))
            if val == "True":
                ret = True
            else:
                ret = False
        else:
            ret = self.defaults[key]
        return ret

    def get_int(self, key):
        """Returns integer configuration key"""
        key = key.lower()
        cdict = self.load()
        if key in cdict:
            val = cdict[key]
            if not val.isdigit():
                raise ValueError("Config key '{}' is no integer!".format(key))
            ret = int(val)
        else:
            raise KeyError("Config key `{}` not set!".format(key))
        return ret

    def get_path(self, name=""):
        """Returns the path for label `name`"""
        cdict = self.load()
        wdkey = "path {}".format(name.lower())

        if wdkey in cdict:
            wd = cdict[wdkey]
        else:
            wd = "./"

        return wd

    def save(self, cdict):
        """Save a settings dictionary into a file"""
        if not self.cfgfile:
            raise SettingsFileError("Settings path not set!")
        skeys = list(cdict.keys())
        skeys.sort()
        outlist = []
        for sk in skeys:
            outlist.append("{} = {}\n".format(sk, cdict[sk]))

        with self.cfgfile.open('w') as fop:
            fop.writelines(outlist)

    def set_bool(self, key, value):
        """Sets boolean key in the settings file"""
        cdict = self.load()
        cdict[key.lower()] = bool(value)
        self.save(cdict)

    def set_int(self, key, value):
        """Sets integer key in the settings file"""
        cdict = self.load()
        cdict[key.lower()] = int(value)
        self.save(cdict)

    def set_path(self, wd, name=""):
        """Set the path in the settings file"""
        cdict = self.load()
        wdkey = "path {}".format(name.lower())
        cdict[wdkey] = wd
        self.save(cdict)


class SettingsFileCache(SettingsFile):
    """A SettingsFile-based data cache"""

    def __init__(self, name):
        directory = appdirs.user_cache_dir()
        super(SettingsFileCache, self).__init__(name=name,
                                                defaults={},
                                                directory=directory)


class SettingsFileError(BaseException):
    pass



import io

import matplotlib as mpl
from matplotlib import pyplot
from matplotlib.figure import Figure
from matplotlib.backends.backend_qt5agg import (
    FigureCanvasQTAgg as FigureCanvas)
import matplotlib.patches as mpatches
from mpl_toolkits.axes_grid1 import make_axes_locatable
import numpy as np


from ..head import custom_widgets


rc = {"color_invalid": "k",
      "color_map": "viridis",
      "color_map_over": "w",
      "color_map_under": "w",
      }


class MPLQMap(object):
    def __init__(self):
        """Matplotlib plot for 2D quantitative map data"""
        # Do not use tight_layout (adjust subplot parameters instead)
        self.figure = Figure(facecolor="none", tight_layout=False)
        # verbose text if no map data is available
        self.no_data_text = self.figure.text(.5, .5,
                                             "no map data available",
                                             horizontalalignment="center")
        self.no_data_text.set_visible(False)
        # main axis
        self.axis_main = self.figure.add_subplot(111)
        cmap = mpl.cm.get_cmap(rc["color_map"])
        cmap.set_over(rc["color_map_over"], 1)
        cmap.set_under(rc["color_map_under"], 1)
        self.plot = self.axis_main.imshow(np.zeros((10, 10)),
                                          interpolation="none",
                                          origin='lower',
                                          cmap=cmap)
        self.canvas = FigureCanvas(self.figure)

        self.axis_main.set_xlabel("position x [µm]")
        self.axis_main.set_ylabel("position y [µm]")
        self.axis_main.set_facecolor(rc["color_invalid"])

        # selection rectangle
        self.select_rect = mpatches.Rectangle(
            (0.1, 0.1), 200, 200, ec="red", fc="none",
            transform=self.axis_main.transData)
        self.axis_main.add_patch(self.select_rect)

        # color bar
        divider = make_axes_locatable(self.axis_main)
        cax = divider.append_axes("right", size="5%", pad=0.05)
        acbar = pyplot.colorbar(self.plot, cax=cax, label="Data")
        acbar.ax.yaxis.set_ticks_position("right")
        acbar.ax.yaxis.set_label_position("right")
        self.colorbar = acbar

        self.lines = []

        # mouse click event
        self.click_callback = None
        self.canvas.mpl_connect('button_press_event', self.on_click)

        # adjust suplot parameters left,bottom,right,top
        self.figure.subplots_adjust(0.2, 0.2, 0.8, 0.95)
        self.canvas.draw()

    def add_toolbar(self, widget):
        """Add mpl toolbar"""
        self.toolbar = custom_widgets.NavigationToolbarQMap(
            self.canvas,
            widget,
            coordinates=True
        )
        self.toolbar.save_data_callback = self.save_data_callback
        self.reset()

    def connect_curve_selection_event(self, callback):
        """Connect mouse click event with callback

        The callback method should accept an index as argument that
        enumerates `self.data_coords`.
        """
        self.click_callback = callback

    def on_click(self, event):
        if (self.click_callback is not None and
            self.qmap_coords is not None and
                event.inaxes is self.axis_main):
            idx = self.set_selection_by_coord(x=event.xdata, y=event.ydata)
            self.click_callback(idx)
        else:
            self.select_rect.set_visible(False)
            self.canvas.draw()

    def reset(self):
        self.colorbar.ax.set_visible(False)
        self.toolbar.setVisible(False)
        self.plot.axes.set_visible(False)
        self.no_data_text.set_visible(True)
        self.qmap = None
        self.qmap_coords = None
        self.qmap_data = None
        self.qmap_shape = (np.nan, np.nan)
        self.dx = 1
        self.dy = 1
        self.reset_lines()

    def reset_lines(self):
        for _ in range(len(self.lines)):
            self.lines.pop(0).remove()

    def save_data_callback(self, filename):
        """Save current image as tsv"""
        with io.open(filename, "wb") as fd:
            np.savetxt(fd, self.qmap_data, delimiter="\t")

    def set_selection_by_coord(self, x, y):
        """Set the position of the red selection frame with coordinates

        Parameters
        ----------
        x,y: float
            The coordinates of the selection. These values do not
            have to be exact.

        Returns
        -------
        index: int
            Index in self.qmap_coords corresponding to the
            selection.

        Notes
        -----
        If the index matches the currently displayed index, then
        the visibility of the selection rectangle patch is toggled.

        See Also
        --------
        set_selection_by_index: Use index instead of coords
        """
        if self.qmap_coords is not None:
            cx = self.qmap_coords[:, 0]
            cy = self.qmap_coords[:, 1]
            index = np.argmin((cx-x)**2 + (cy-y)**2)
            self.set_selection_by_index(index=index)
            return index
        else:
            return None

    def set_selection_by_index(self, index):
        """Set the position of the red selection frame with index

        Parameters
        ----------
        index: int
            Index in self.qmap_coords corresponding to the
            selection.

        Notes
        -----
        If the index matches the currently displayed index, then
        the visibility of the selection rectangle patch is toggled.

        See Also
        --------
        set_selection_by_coord: Use coordinates instead of index
        """
        if self.qmap_coords is not None:
            cx = self.qmap_coords[:, 0]
            cy = self.qmap_coords[:, 1]
            newxy = (cx[index]-self.dx/2, cy[index]-self.dy/2)
            self.select_rect.set_visible(True)
            self.select_rect.set_xy(newxy)
            self.select_rect.set_width(self.dx)
            self.select_rect.set_height(self.dy)
            self.canvas.draw()

    def update(self, qmap, feature, cmap="viridis", vmin=None, vmax=None):
        """Update the map tab plot data

        Parameters
        ----------
        qmap: nanite.QMap
            Qmap data
        vmin, vmax: float
            Data range for plotting
        label: str
            QMap colorbar data label

        Notes
        -----
        If `coords` or `qmap_data` is set to none, then a message will
        be displayed in the plot stating that no map data is available.
        """
        prev_data = self.qmap_data

        qmap_data = qmap.get_qmap(feature=feature, qmap_only=True)
        qmap_coords = qmap.get_coords(which="um")
        qmap_coords_px = qmap.get_coords(which="px")
        shape = qmap_data.shape
        extent = qmap.extent
        dx = (extent[1] - extent[0])/shape[0]
        dy = (extent[3] - extent[2])/shape[1]

        # TODO:
        # - only update the plot if vmin/vmax or qmap_data as changed

        if vmin == vmax:
            vmin = vmax = None

        if vmin is None:
            vmin = np.nanmin(qmap_data)
        if vmax is None:
            vmax = np.nanmax(qmap_data)
        self.plot.set_clim(vmin=vmin, vmax=vmax)

        self.plot.set_cmap(cmap)

        if (prev_data is None
                or not np.allclose(qmap_data, prev_data, equal_nan=True)):
            # visibility of plot elements
            self.colorbar.ax.set_visible(True)
            self.toolbar.setVisible(True)
            self.plot.axes.set_visible(True)
            self.no_data_text.set_visible(False)

            # set plot data
            self.plot.set_data(qmap_data)
            self.plot.set_extent(extent)
            self.colorbar.set_label(feature)

            # set invalid elements
            self.reset_lines()
            xm, ym = np.meshgrid(range(shape[0]),
                                 range(shape[1]))
            for xi, yi in zip(xm.flat, ym.flat):
                if np.isnan(qmap_data[yi, xi]):
                    xv = qmap_coords[0][0] + xi * dx
                    yv = qmap_coords[0][1] + yi * dy
                    if (np.sum(np.all(np.array([xi, yi]) == p)
                               for p in qmap_coords_px)):
                        color = "#14571A"
                    else:
                        color = "#571714"
                    self.lines.append(
                        self.axis_main.plot([xv-dx*.4, xv+dx*.4],
                                            [yv-dy*.4, yv+dy*.4],
                                            color=color,
                                            lw=1)[0]
                        )

            # common variables
            self.dx = dx
            self.dy = dy
            self.qmap = qmap
            self.qmap_coords = qmap_coords
            self.qmap_data = qmap_data

        self.canvas.draw()



"""Test of data set functionalities"""
import pathlib

from PyQt5 import QtWidgets

import nanite.model as nmodel
import pyjibe.head


here = pathlib.Path(__file__).parent
jpkfile = here / "data" / "spot3-0192.jpk-force"


def cleanup_autosave(jpkfile):
    """Remove autosave files"""
    path = jpkfile.parent
    files = path.glob("*.tsv")
    files = [f for f in files if f.name.startswith("pyjibe_")]
    [f.unlink() for f in files]


def test_remember_initial_params(qtbot):
    cleanup_autosave(jpkfile)
    main_window = pyjibe.head.PyJibe()
    main_window.load_data(files=[jpkfile, jpkfile])
    war = main_window.subwindow_data[0]
    # clear data
    war.list_preproc_applied.clear()
    war.cb_autosave.setChecked(0)
    # perform simple filter
    item = QtWidgets.QListWidgetItem()
    item.setText("compute_tip_position")
    war.list_preproc_applied.addItem(item)
    # perform fitting with standard parameters
    # set initial parameters in user interface
    itab = war.table_parameters_initial
    # disable weighting
    war.cb_weight_cp.setCheckState(0)
    # enable fitting of force offset
    itab.item(4, 0).setCheckState(0)
    # set better value for contact point
    itab.item(3, 1).setText(str(18000))
    # change standard tip radius from 10 to 5
    assert float(itab.item(1, 1).text()) == 10
    itab.item(1, 1).setText(str(5))
    cl1 = war.list_curves.currentItem()
    cl2 = war.list_curves.itemBelow(cl1)
    war.list_curves.setCurrentItem(cl2)
    assert float(itab.item(1, 1).text()) == 5
    cleanup_autosave(jpkfile)


def test_change_model_keep_parms(qtbot):
    cleanup_autosave(jpkfile)
    main_window = pyjibe.head.PyJibe()
    main_window.load_data(files=[jpkfile, jpkfile])
    war = main_window.subwindow_data[0]
    # clear data
    war.list_preproc_applied.clear()
    war.cb_autosave.setChecked(0)
    # perform simple filter
    item = QtWidgets.QListWidgetItem()
    item.setText("compute_tip_position")
    war.list_preproc_applied.addItem(item)
    # perform fitting with standard parameters
    # set initial parameters in user interface
    itab = war.table_parameters_initial
    # set value for contact point
    itab.item(3, 1).setText(str(12345))
    # change the model to pyramidal
    pyr_name = nmodel.model_hertz_three_sided_pyramid.model_name
    pyr_idx = war.cb_model.findText(pyr_name)
    war.cb_model.setCurrentIndex(pyr_idx)
    # check that contact point is still the same
    assert float(itab.item(3, 1).text()) == 12345
    cleanup_autosave(jpkfile)



import hashlib
import os
import time

import numpy as np
from PyQt5 import QtCore, QtWidgets, QtGui

import nanite
import nanite.model as nmodel
import nanite.read as nread

from .. import colormap
from .. import units

from .base import UiForceDistanceBase, DlgAutosave
from . import export
from . import user_rating


class UiForceDistance(UiForceDistanceBase):
    def __init__(self, parent_widget):
        """Base class derived from Qt designer

        To reduce the number of lines in a file, the UI for
        force-indentation measurements is split into two classes:
        - `UiIndentationBase` completes the (dynamic) design of the UI and
          provides convenience properties.
        - `UiIndentation` contains logic parts of the code that are not
          part of the UI.
        """
        super(UiForceDistance, self).__init__(parent_widget)
        # Random string for identification of autosaved results
        self._autosave_randstr = hashlib.md5(time.ctime().encode("utf-8")
                                             ).hexdigest()[:5]
        # Initialize `override` parameter:
        # -1: not decided
        #  0: do not override existing files
        #  1: override existing files
        #  2: create additional file with date
        self._autosave_override = -1
        # Filenames that were created by this instance
        self._autosave_original_files = []

    def add_files(self, files):
        """ Populate self.data_set and display the first curve """
        # The `mult` parameter is used to chunk the progress bar,
        # because we cannot use floats with `QProgressDialog`, but
        # we want to be able to show progress for files that contain
        # multiple force curves.
        mult = 100
        bar = QtWidgets.QProgressDialog("Loading data files...",
                                        "Stop", 1, len(files)*mult)
        bar.setWindowTitle("Loading data files")
        bar.setMinimumDuration(1000)
        for ii, f in enumerate(files):
            label = "Loading file\n{}".format(os.path.basename(f))
            bar.setLabelText(label)

            def callback(partial):
                """Call back method for a progress dialog

                This method updates `bar` with the current file index
                `ii` and a partial file value `partial`.

                Parameters
                ----------
                partial: float in [0,1]
                    The progress for a single file
                """
                bar.setValue(int((ii+partial)*mult))
                QtCore.QCoreApplication.instance().processEvents()
                if bar.wasCanceled():
                    # Raise a custom `AbortProgress` error, such that
                    # we can exit the parent for-loop.
                    raise AbortProgress
            try:
                grp = nanite.IndentationGroup(f, callback=callback)
                callback(1)
            except nread.read_jpk_meta.ReadJPKMetaKeyError:
                # ignore callibration curves
                callback(1)
                continue
            except AbortProgress:
                # The custom error `AbortProgress` was called, because
                # the user wants to stop loading the data.
                break
            self.data_set += grp
        bar.reset()
        bar.close()
        self.curve_list_setup()
        # Select first item
        it = self.list_curves.topLevelItem(0)
        self.list_curves.setCurrentItem(it)

    def autosave(self, fdist):
        """Performs autosaving for all files"""
        if (self.cb_autosave.checkState() == 2 and
            fdist.fit_properties and
                fdist.fit_properties["success"]):
            # Determine the directory of the current curve
            adir = os.path.dirname(fdist.path)
            model_key = fdist.fit_properties["model_key"]
            # Determine all other curves with the same path
            exp_curv = []
            for ii, ar in enumerate(self.data_set):
                it = self.list_curves.topLevelItem(ii)
                if (  # same directory
                    os.path.dirname(ar.path) == adir and
                    # fdist was fitted
                    ar.fit_properties and
                    # fit was successful
                    ar.fit_properties["success"] and
                    # fdist was fitted with same model
                    ar.fit_properties["model_key"] == model_key and
                    # user selected curve for export ("use")
                    it.checkState(3) == 2
                ):
                    exp_curv.append(ar)
            # The file to export
            fname = os.path.join(adir, "pyjibe_fit_results_leaf.tsv")

            # Only export if we have curves to export
            if exp_curv:
                if (os.path.exists(fname) and
                        fname not in self._autosave_original_files):
                    # File already exists
                    oride = self._autosave_override
                    if oride == -1:
                        # Ask user what to do
                        dlgwin = QtWidgets.QDialog()
                        dlgwin.setWindowModality(QtCore.Qt.ApplicationModal)
                        dlgui = DlgAutosave()
                        dlgui.setupUi(dlgwin)
                        if dlgwin.exec_():
                            if dlgui.btn_nothing.isChecked():
                                oride = 0
                            elif dlgui.btn_override.isChecked():
                                oride = 1
                            elif dlgui.btn_new.isChecked():
                                oride = 2
                            if dlgui.cb_remember.isChecked():
                                self._autosave_override = oride
                            if dlgui.cb_disableauto.isChecked():
                                self.cb_autosave.setChecked(0)
                    if oride == 0:
                        # Do not override
                        return
                    elif oride == 1:
                        # Override existing file
                        pass
                    elif oride == 2:
                        # Create new filename
                        now = time.strftime("%Y-%m-%d")
                        newbase = "pyjibe_fit_results_leaf_{}_{}.tsv".format(
                            now,
                            self._autosave_randstr
                        )
                        fname = os.path.join(adir, newbase)
                # Export data
                ratings = self.rate_data(exp_curv)
                export.save_tsv_approach_retract(filename=fname,
                                                 fdist_list=exp_curv,
                                                 ratings=ratings)
                self._autosave_original_files.append(fname)

    def curve_list_update(self, item=None):
        """Update the curve list display with all ratings"""
        if item is None:
            indices = np.arange(len(self.data_set))
        else:
            indices = [item]
        for ii in indices:
            ar = self.data_set[ii]
            rating = self.rate_data(ar)
            it = self.list_curves.topLevelItem(ii)
            it.setText(2, "{:.1f}".format(rating))
            cm = colormap.cm_rating
            color = np.array(cm(rating/10))*255
            it.setBackground(2, QtGui.QColor(*color))

    def fit_apply_preprocessing(self, fdist):
        """Apply the preprocessing steps if required"""
        # Note: Preprocessing is cached once in `fdist`.
        # Thus calling this method a second time without any
        # change in the GUI is free.
        num = self.list_preproc_applied.count()
        preprocessing = []
        for ii in range(num):
            item = self.list_preproc_applied.item(ii)
            preprocessing.append(item.text())
        # Perform preprocessing
        fdist.apply_preprocessing(preprocessing)

    def fit_approach_retract(self, fdist, update_ui=True):
        """Perform preprocessing and fit data

        Parameters
        ----------
        fdist: afmlib.indentaion.Indentaion
            Indentation curve to fit.
        update_ui: bool
            Update the user interface after fitting,
            i.e. displaying the results in
            `self.table_parameters_fitted`.
        """
        # segment
        segment = self.cb_segment.currentText().lower()
        # x axis
        x_axis = self.cb_xaxis.currentText()
        # y axis
        y_axis = self.cb_yaxis.currentText()
        # Get model key from dropdown list
        model = nmodel.get_model_by_name(self.cb_model.currentText())
        model_key = model.model_key
        # Determine range type
        if self.cb_range_type.currentText() == "absolute":
            range_type = "absolute"
        else:
            range_type = "relative cp"
        # Determine range
        range_x = [self.sp_range_1.value() * units.scales["µ"],
                   self.sp_range_2.value() * units.scales["µ"]]
        # Determine if we want to weight the contact point
        self.on_update_weights(on_params_init=False)
        if self.cb_weight_cp.checkState() == 2:
            weight_cp = self.sp_weight_cp_um.value() * units.scales["µ"]
        else:
            weight_cp = False
        # Determine if we want to autodetect the optimal indentation depth
        if self.cb_delta_select_1.currentIndex() == 2:
            optimal_fit_edelta = True
        else:
            optimal_fit_edelta = False
        # number of samples for edelta plot
        optimal_fit_num_samples = self.sp_delta_num_samples.value()
        # fit parameters
        params = self.fit_parameters()
        # Perform fitting
        fdist.fit_model(model_key=model_key,
                        params_initial=params,
                        range_x=range_x,
                        range_type=range_type,
                        x_axis=x_axis,
                        y_axis=y_axis,
                        weight_cp=weight_cp,
                        segment=segment,
                        optimal_fit_edelta=optimal_fit_edelta,
                        optimal_fit_num_samples=optimal_fit_num_samples,
                        )
        ftab = self.table_parameters_fitted
        if fdist.fit_properties["success"]:
            # Perform automatic saving of results
            self.autosave(fdist)
            # Display automatically detected optimal indentation depth
            if optimal_fit_edelta:
                # Set guessed indentation depth in GUI
                val = fdist.fit_properties["optimal_fit_delta"]
                self.sp_range_1.setValue(val/units.scales["µ"])
            if update_ui:
                # Display results in `self.table_parameters_fitted`
                fitpar = fdist.fit_properties["params_fitted"]
                varps = [p[1] for p in fitpar.items() if p[1].vary]
                self.assert_parameter_table_rows(ftab, len(varps))
                for ii, p in enumerate(varps):
                    # Get the human readable name of the parameter
                    name = model.parameter_keys.index(p.name)
                    hrnam = model.parameter_names[name]
                    # Determine unit scale, e.g. 1e6 [sic] for µm
                    scale = units.hrscale(hrnam)
                    ftab.verticalHeaderItem(ii).setText(units.hrscname(hrnam))
                    ftab.item(ii, 0).setText("{:.3f}".format(p.value*scale))
        else:
            if update_ui:
                inipar = fdist.fit_properties["params_initial"]
                varps = [p[1] for p in inipar.items() if p[1].vary]
                self.assert_parameter_table_rows(ftab, len(varps))
                for ii, p in enumerate(varps):
                    # Get the human readable name of the parameter
                    name = model.parameter_keys.index(p.name)
                    hrnam = model.parameter_names[name]
                    ftab.verticalHeaderItem(ii).setText(units.hrscname(hrnam))
                    ftab.item(ii, 0).setText("nan")

    def info_update(self, fdist):
        """Updates the info tab"""
        text = []
        text.append("filename: {}".format(fdist.path))
        text.append("position index/enum: {}".format(fdist.enum))
        text.append("")
        keys = list(fdist.metadata.keys())
        keys.sort()
        for k in keys:
            text.append("{}: {}".format(k, fdist.metadata[k]))
        textstring = "\n".join(text)
        self.info_text.setPlainText(textstring)

    def mpl_curve_update(self, fdist):
        """Update the force-indentation curve"""
        autoscale_x = self.cb_mpl_rescale_plot_x.checkState() == 2
        autoscale_y = self.cb_mpl_rescale_plot_y.checkState() == 2
        if autoscale_x:
            rescale_x = None
        else:
            rescale_x = (self.cb_mpl_rescale_plot_x_min.value(),
                         self.cb_mpl_rescale_plot_x_max.value())

        if autoscale_y:
            rescale_y = None
        else:
            rescale_y = (self.cb_mpl_rescale_plot_y_min.value(),
                         self.cb_mpl_rescale_plot_y_max.value())

        self.mpl_curve.update(fdist,
                              rescale_x=rescale_x,
                              rescale_y=rescale_y)

    def mpl_edelta_update(self):
        """Update the E(delta) plot"""
        if self.tabs.currentWidget() == self.tab_edelta:
            fdist = self.current_curve
            delta_opt = self.sp_range_1.value()
            # Update slider range
            xaxis = self.cb_xaxis.currentText()
            segment = self.cb_segment.currentText().lower()
            segment_bool = segment == "retract"
            segid = (fdist["segment"] == segment_bool).as_matrix()
            xdata = fdist[xaxis].as_matrix()
            xscale = units.hrscale(xaxis)
            minx = np.min(xdata[segid])*xscale
            self.delta_slider.blockSignals(True)
            self.delta_slider.setMinimum(minx)
            self.delta_slider.setValue(delta_opt)
            self.delta_slider.blockSignals(False)
            # Update E(delta) plot
            self.fit_approach_retract(fdist)
            self.mpl_edelta.update(fdist, delta_opt)

    def mpl_qmap_update(self):
        fdist = self.current_curve
        # Only update if we are on the right tab
        if self.tabs.currentWidget() == self.tab_qmap:
            # Build list of possible selections
            selist = nanite.qmap.available_features

            # Get plotting parameter and check if it makes sense
            feature = self.qmap_data_cb.currentText()
            if not feature or feature not in selist:
                # Use a default plotting map
                feature = "data min height"

            # Make sure that we have a valid property to plot
            assert feature in selist

            # Update dropdown menu with possible selections
            # disable signals while updating the combobox
            self.qmap_data_cb.blockSignals(True)
            # remove all items
            for _i in range(self.qmap_data_cb.count()):
                self.qmap_data_cb.removeItem(0)
            # add new items
            for item in selist:
                self.qmap_data_cb.addItem(item)
            self.qmap_data_cb.setCurrentIndex(selist.index(feature))
            self.qmap_data_cb.blockSignals(False)

            # Get all selected curves with the same path
            curves = self.selected_curves.subgroup_with_path(fdist.path)

            if len(curves) > 1:
                # Get map data
                qmap = nanite.QMap(curves)
                # update plot
                self.mpl_qmap.update(qmap=qmap,
                                     feature=feature,
                                     cmap=self.qmpa_cmap_cb.currentText(),
                                     vmin=self.qmap_sp_range1.value(),
                                     vmax=self.qmap_sp_range2.value())
                self.mpl_qmap.set_selection_by_index(curves.index(fdist))
            else:
                self.mpl_qmap.reset()

    def on_curve_list(self):
        """Called when a new curve is selected"""
        fdist = self.current_curve
        idx = self.current_index
        # perform preprocessing
        self.fit_apply_preprocessing(fdist)
        # update user interface with initial parameters
        self.fit_update_parameters(fdist)
        # fit data
        self.fit_approach_retract(fdist)
        # set plot data (time consuming)
        self.mpl_curve_update(fdist)
        # update info
        self.info_update(fdist)
        # Display new rating
        self.curve_list_update(item=idx)
        # Display map
        self.mpl_qmap_update()
        # Display edelta
        self.mpl_edelta_update()

    def on_curve_list_item_changed(self, item):
        """An item in the curve list was changed

        - update the qmap
        - autosave the results
        """
        # The data set index:
        idx = item.row()
        fdist = self.data_set[idx]
        if item.column() == 3:
            # The checkbox has been changed
            self.mpl_qmap_update()
            self.autosave(fdist)

    def on_model(self):
        """Called when the fitting model is changed"""
        # The difference to "on_params_init" is that we
        # have to `fit_update_parameters` in order to display
        # potential new parameter names of the new model.
        fdist = self.current_curve
        self.fit_apply_preprocessing(fdist)
        self.fit_update_parameters(fdist)
        self.fit_approach_retract(fdist)
        self.mpl_curve_update(fdist)
        self.curve_list_update()
        self.mpl_qmap_update()

    def on_params_init(self):
        """Called when the initial parameters are changed"""
        fdist = self.current_curve
        idx = self.current_index
        self.fit_apply_preprocessing(fdist)
        self.fit_approach_retract(fdist)
        self.mpl_curve_update(fdist)
        self.curve_list_update(item=idx)
        self.mpl_qmap_update()

    def on_tab_changed(self, index):
        """Called when the tab on the right hand is changed"""
        if hasattr(self, "user_tab_selected"):
            prevtab = self.user_tab_selected
        else:
            prevtab = self.tabs.currentWidget()

        curtab = self.tabs.currentWidget()
        if curtab == self.tab_fit and prevtab == self.tab_preprocess:
            self.on_params_init()
        elif curtab == self.tab_qmap:
            # Redraw the current map
            self.mpl_qmap_update()
        elif curtab == self.tab_edelta:
            # Compute edelta plot
            self.mpl_edelta_update()

        self.user_tab_selected = curtab

    def on_user_rate(self):
        cont = QtWidgets.QFileDialog.getSaveFileName(
            parent=None,
            caption="Please select a rating container",
            directory="",
            filter="Rating containers (*.h5)",
            options=QtWidgets.QFileDialog.DontConfirmOverwrite
                    | QtWidgets.QFileDialog.DontUseNativeDialog)

        path = cont[0]
        if path:
            rt = user_rating.Rater(fdui=self, path=path)
            self.curve_rater = rt
            rt.show()


class AbortProgress(BaseException):
    pass



import pathlib

from pyjibe import settings


def test_cfg_basic():
    cfg = settings.SettingsFile()
    wd = pathlib.Path(".").resolve()
    cfg.set_path(str(wd.parent), "Peter")

    assert wd.parent == pathlib.Path(cfg.get_path("Peter")).resolve()


if __name__ == "__main__":
    # Run all tests
    loc = locals()
    for key in list(loc.keys()):
        if key.startswith("test_") and hasattr(loc[key], "__call__"):
            loc[key]()



import collections
import inspect
import io
import pkg_resources

import numpy as np
from PyQt5 import uic, QtCore, QtWidgets

import nanite
import nanite.fit as nfit
import nanite.model as nmodel
import nanite.indent as nindent
import nanite.preproc as npreproc

from .. import units

from . import export
from .mpl_indent import MPLIndentation
from .mpl_edelta import MPLEDelta
from .mpl_qmap import MPLQMap


RATING_SCHEMES = collections.OrderedDict()
RATING_SCHEMES["Default (zef18 & Extra Trees)"] = ["zef18", "Extra Trees"]
RATING_SCHEMES["Disabled"] = ["none", "none"]


# load QWidget from ui file
ui_path = pkg_resources.resource_filename("pyjibe.fd",
                                          "base_design.ui")
UiForceDistanceCore = uic.loadUiType(ui_path)[0]

dlg_autosave_path = pkg_resources.resource_filename("pyjibe.fd",
                                                    "dlg_autosave_design.ui")
DlgAutosave = uic.loadUiType(dlg_autosave_path)[0]


class UiForceDistanceBase(UiForceDistanceCore):
    _instance_counter = 0

    def __init__(self, parent_widget):
        """Base class derived from Qt designer

        To reduce the number of lines in a file, the UI for
        force-indentation measurements is split into two classes:
        - `UiIndentationBase` completes the (dynamic) design of the UI and
          provides convenience properties.
        - `UiIndentation` contains logic parts of the code that are not
          part of the UI.
        """
        super(UiForceDistanceCore, self).__init__()
        self.setupUi(parent_widget)
        self.parent_widget = parent_widget

        UiForceDistanceBase._instance_counter += 1
        title = "{} #{}".format(self.parent_widget.windowTitle(),
                                self._instance_counter)
        self.parent_widget.setWindowTitle(title)

        self.mpl_curve_setup()
        self.mpl_edelta_setup()
        self.mpl_qmap_setup()
        self.data_set = nanite.IndentationGroup()

        # initial values, sources, drains for indentation depth
        self.indentation_depth_setup()

        # preprocessing setup
        self.preproc_setup()

        # fitting setup
        self.fit_setup()

        self.tabs.currentChanged.connect(self.on_tab_changed)
        self.signal_slot(True)
        self.btn_rater.clicked.connect(self.on_user_rate)

    def assert_parameter_table_rows(self, table, rows, cb_first=False):
        """Make sure a QTableWidget has enough rows

        Parameters
        ----------
        table: instance of QtWidgets.QTableWidget
            The table that will be edited
        rows: int
            Number of rows
        cb_first: bool
            Set first column values as check-boxes.
            (This is used for fixing fit parameters in PyJibe)
        """
        table.setRowCount(rows)
        cols = table.columnCount()
        for rr in range(rows):
            hitem = QtWidgets.QTableWidgetItem()
            table.setVerticalHeaderItem(rr, hitem)
            for cc in range(cols):
                item = QtWidgets.QTableWidgetItem()
                if cc == 0 and cb_first:
                    item.setFlags(QtCore.Qt.ItemIsUserCheckable |
                                  QtCore.Qt.ItemIsEnabled)
                    item.setCheckState(QtCore.Qt.Unchecked)
                table.setItem(rr, cc, item)

    @property
    def current_curve(self):
        idx = self.current_index
        fdist = self.data_set[idx]
        return fdist

    @property
    def current_index(self):
        item = self.list_curves.currentItem()
        idx = self.list_curves.indexOfTopLevelItem(item)
        return idx

    def curve_list_setup(self):
        """Add items to the tree widget"""
        header = self.list_curves.header()
        header.setStretchLastSection(False)
        header.setSectionResizeMode(0, QtWidgets.QHeaderView.Stretch)
        self.list_curves.setColumnWidth(1, 70)
        self.list_curves.setColumnWidth(2, 70)
        self.list_curves.setColumnWidth(3, 40)
        for ar in self.data_set:
            it = QtWidgets.QTreeWidgetItem(self.list_curves,
                                           ["..."+str(ar.path)[-62:],
                                            str(ar.enum),
                                            "{:.1f}".format(-1)])
            self.list_curves.addTopLevelItem(it)
            it.setCheckState(3, QtCore.Qt.Checked)
        # Connect signals:
        # Selection of curves
        self.list_curves.itemSelectionChanged.connect(self.on_curve_list)
        self.list_curves.model().dataChanged.connect(
            self.on_curve_list_item_changed)

    def fit_parameters(self):
        """Returns parameters currently set in the GUI

        The parameter data is read out from
        `self.table_parameters_initial`. If the model that was
        previously set does not match the current model, only the
        parameters with the same name (existing in both models) are
        updated.
        """
        # Get model key from dropdown list
        model = nmodel.get_model_by_name(self.cb_model.currentText())
        model_key = model.model_key
        # Get parameters from `self.table_parameters_initial`
        model = nmodel.models_available[model_key]
        params = model.get_parameter_defaults()
        itab = self.table_parameters_initial
        if itab.verticalHeaderItem(0):
            # Only update if there is already something set
            for ii, key in enumerate(list(params.keys())):
                p = params[key]
                # Determine the scale for which we need to revert
                # the multiplications.
                # performed in `fit_update_parameters`.
                hrname = model.parameter_names[ii]
                scale = units.hrscale(hrname)
                for rr in range(itab.rowCount()):
                    # search for a row matching the parameter `p`
                    if itab.verticalHeaderItem(rr).text().startswith(hrname):
                        # update parameter `p`
                        state = itab.item(rr, 0).checkState()
                        if state == QtCore.Qt.Unchecked:
                            p.vary = True
                        else:
                            p.vary = False
                        p.set(float(itab.item(rr, 1).text())/scale)
                        p.min = float(itab.item(rr, 2).text())/scale
                        p.max = float(itab.item(rr, 3).text())/scale
                        break
        return params

    def fit_setup(self):
        """Setup the fitting tab"""
        id_para = 0
        # Model selection
        models_av = list(nmodel.models_available.keys())
        models_av.sort(key=lambda x: nmodel.models_available[x].model_name)
        for ii, key in enumerate(models_av):
            model = nmodel.models_available[key]
            self.cb_model.addItem(model.model_name)
            if key == "hertz_para":
                id_para = ii
        # Setz Hertz Parabolic as default
        self.cb_model.setCurrentIndex(id_para)
        # Axis selection
        self.cb_xaxis.addItem("tip position")
        self.cb_yaxis.addItem("force")

    def fit_update_parameters(self, fdist):
        """Update the initial and fitting parameters"""
        model = nmodel.get_model_by_name(self.cb_model.currentText())
        model_key = model.model_key
        if ("params_initial" in fdist.fit_properties and
            "model_key" in fdist.fit_properties and
                fdist.fit_properties["model_key"] == model_key):
            # set the parameters of the previous fit
            params = fdist.fit_properties["params_initial"]
        else:
            # use the initial model parameters
            params = self.fit_parameters()
        # parameter table
        itab = self.table_parameters_initial

        itab.setColumnWidth(0, 30)
        itab.setColumnWidth(1, 100)
        itab.setColumnWidth(2, 70)
        itab.setColumnWidth(3, 70)
        itab.blockSignals(True)

        self.assert_parameter_table_rows(itab, len(params), cb_first=True)
        for ii, key in enumerate(list(params.keys())):
            p = params[key]
            # Get the human readable name of the parameter
            hrname = model.parameter_names[ii]
            # Determine unit scale, e.g. 1e6 [sic] for µm
            scale = units.hrscale(hrname)
            itab.verticalHeaderItem(ii).setText(units.hrscname(hrname))
            if p.vary:
                state = QtCore.Qt.Unchecked
            else:
                state = QtCore.Qt.Checked
            itab.item(ii, 0).setCheckState(state)
            itab.item(ii, 1).setText("{:.2f}".format(p.value*scale))
            itab.item(ii, 2).setText(str(p.min*scale))
            itab.item(ii, 3).setText(str(p.max*scale))

        itab.blockSignals(False)

        # indentation depth
        if self.cb_delta_select_1.currentIndex() == 1:
            # Set indentation depth individually
            if fdist in self._indentation_depth_individual:
                value = self._indentation_depth_individual[fdist]
                self.delta_spin.setValue(value)
            else:
                value = self.delta_spin.value()
                self._indentation_depth_individual[fdist] = value

    @staticmethod
    def get_export_choices():
        """Choices for file menu export

        Returns
        -------
        choices: list
            [[menu label 1, method name 1],
             ...
             ]
        """
        choices = [["fit results", "on_export_fit_results"],
                   ["E(δ) curves", "on_export_edelta"]
                   ]
        return choices

    def indentation_depth_setup(self):
        """Initiate ranges (spin/slider) that allow inf values"""
        # Initialize individual indentation depth dictionary
        self._indentation_depth_individual = {}
        # Left
        self.sp_range_1.setValue(-np.inf)
        # Right
        self.sp_range_2.setValue(np.inf)
        # Left on edelta plot
        self.delta_spin.setValue(-np.inf)
        # Slider
        self.delta_slider.setMaximum(0)
        # Signals
        self.sp_range_1.valueChanged["double"].connect(
            self.on_delta_change_spin)
        self.delta_spin.valueChanged["double"].connect(
            self.on_delta_change_spin)
        self.delta_slider.sliderReleasedFloat["double"].connect(
            self.on_delta_change_spin)
        self.delta_slider.valueChangedFloat["double"].connect(
            self.mpl_edelta.update_delta)
        self.sp_range_2.valueChanged.connect(self.on_params_init)
        self.cb_delta_select_1.currentIndexChanged['int'].connect(
            self.on_delta_select)
        self.delta_btn_guess.clicked.connect(self.on_delta_guess)

    def mpl_curve_setup(self):
        """Setup the matplotlib interface for approach retract plotting"""
        self.mpl_curve = MPLIndentation()
        self.mpl_curve.add_toolbar(self.mplwindow)
        self.mplvl.addWidget(self.mpl_curve.canvas)
        self.mplvl.addWidget(self.mpl_curve.toolbar)
        self.cb_mpl_rescale_plot_x.stateChanged.connect(
            self.on_mpl_curve_update)
        self.cb_mpl_rescale_plot_x_min.valueChanged.connect(
            self.on_mpl_curve_update)
        self.cb_mpl_rescale_plot_x_max.valueChanged.connect(
            self.on_mpl_curve_update)
        self.cb_mpl_rescale_plot_y.stateChanged.connect(
            self.on_mpl_curve_update)
        self.cb_mpl_rescale_plot_y_min.valueChanged.connect(
            self.on_mpl_curve_update)
        self.cb_mpl_rescale_plot_y_max.valueChanged.connect(
            self.on_mpl_curve_update)

    def mpl_edelta_setup(self):
        """Setup the matplotlib interface for E(delta) plotting"""
        self.mpl_edelta = MPLEDelta()
        self.mpl_edelta.add_toolbar(self.edelta_mplwidget)
        self.edelta_mpllayout.addWidget(self.mpl_edelta.canvas)
        self.edelta_mpllayout.addWidget(self.mpl_edelta.toolbar)

    def mpl_qmap_setup(self):
        """Setup the matplotlib interface for 2D map plotting"""
        self.mpl_qmap = MPLQMap()
        self.mpl_qmap.add_toolbar(self.qmap_mplwidget)
        self.qmap_mpllayout.addWidget(self.mpl_qmap.canvas)
        self.qmap_mpllayout.addWidget(self.mpl_qmap.toolbar)
        # set colormaps
        cmaps = ["viridis", "plasma", "afmhot", "seismic"]
        for cm in cmaps:
            self.qmpa_cmap_cb.addItem(cm)
        self.qmpa_cmap_cb.setCurrentIndex(0)
        self.qmap_data_cb.currentIndexChanged.connect(
            self.on_qmap_data_changed)
        self.qmpa_cmap_cb.currentIndexChanged.connect(
            self.on_qmap_cmap_changed)
        self.qmap_sp_range1.valueChanged.connect(self.on_qmap_min_max_changed)
        self.qmap_sp_range2.valueChanged.connect(self.on_qmap_min_max_changed)
        self.mpl_qmap.connect_curve_selection_event(self.on_qmap_selection)

    def on_mpl_curve_update(self):
        fdist = self.current_curve
        self.mpl_curve_update(fdist)

    def on_qmap_cmap_changed(self):
        """colormap selection changed"""
        self.mpl_qmap_update()

    def on_qmap_data_changed(self):
        """data column selection changed"""
        # set previous spin control values if existent
        self.qmap_sp_range1.blockSignals(True)
        self.qmap_sp_range2.blockSignals(True)
        if hasattr(self, "_cache_qmap_spin_ctl"):
            data = self.qmap_data_cb.currentIndex()
            if data in self._cache_qmap_spin_ctl:
                vmin, vmax = self._cache_qmap_spin_ctl[data]
            else:
                vmin = vmax = 0
            self.qmap_sp_range1.setValue(vmin)
            self.qmap_sp_range2.setValue(vmax)
        self.qmap_sp_range1.blockSignals(False)
        self.qmap_sp_range2.blockSignals(False)
        self.mpl_qmap_update()

    def on_qmap_min_max_changed(self):
        """min or max spin controls changed"""
        # store spin control values for data column
        vmin = self.qmap_sp_range1.value()
        vmax = self.qmap_sp_range2.value()
        data = self.qmap_data_cb.currentIndex()
        if not hasattr(self, "_cache_qmap_spin_ctl"):
            self._cache_qmap_spin_ctl = {}
        self._cache_qmap_spin_ctl[data] = (vmin, vmax)
        self.mpl_qmap_update()

    def on_qmap_selection(self, idx):
        """Show the curve indexed in the current qmap"""
        # Get the qmap name
        cc = self.current_curve
        # idx is `enum` and curves are sorted
        curves = [ci for ci in self.data_set if ci.path == cc.path]
        fdist = curves[idx]
        idcurve = self.data_set.index(fdist)
        item = self.list_curves.topLevelItem(idcurve)
        self.list_curves.setCurrentItem(item)

    def on_delta_guess(self):
        """Guess the optimal indentation depth for the current curve"""
        fdist = self.current_curve
        value = fdist.estimate_optimal_mindelta()
        value /= units.scales["µ"]
        self.delta_spin.setValue(value)

    def on_delta_select(self, value):
        """The user selected a method for indentation depth determination

        This method is called when `cb_delta_select_1` or (indirectly)
        `cb_delta_select_2` are changed.

        At the end, `on_params_init` is called, to redo the fit. After
        fitting, the indentation depth (`sp_range_1`) will be set by
        `fit_approach_retract`, if `value==2`.

        Parameters
        ----------
        value: int
            The selected index of the dropdown widget:
            - 0: Set indentation depth globally
            - 1: Set indentation depth individually
            - 2: Guess optimal indentation depth
        """
        # These widgets are disabled when the user wants
        # PyJibe to guess the indentation depth:
        global_disable = [self.sp_range_1]
        # These widgets are enabled when the user wants to
        # select an indentation depth for each curve:
        local_enable = [self.delta_label,
                        self.delta_spin,
                        self.delta_slider,
                        self.delta_btn_guess]

        if value == 0:
            # Set indentation depth globally
            [item.setEnabled(True) for item in global_disable]
            [item.setEnabled(False) for item in local_enable]
            self.cb_range_type.setEnabled(True)
        elif value == 1:
            # Set indentation depth individually
            [item.setEnabled(True) for item in global_disable]
            [item.setEnabled(True) for item in local_enable]
            self.cb_range_type.setEnabled(False)
            self.cb_range_type.setCurrentText("absolute")
            # Set/get indentation depth individually
            fdist = self.current_curve
            if fdist in self._indentation_depth_individual:
                value = self._indentation_depth_individual[fdist]
            else:
                value = self.delta_spin.value()
            self.delta_spin.setValue(value)

        elif value == 2:
            # Guess optimal indentation depth
            [item.setEnabled(False) for item in global_disable]
            [item.setEnabled(False) for item in local_enable]
            self.cb_range_type.setEnabled(False)
            self.cb_range_type.setCurrentText("absolute")

        self.on_params_init()

    def on_delta_change_spin(self, value):
        """Indentation depth spin control value changed"""
        # Update all controls
        self.sp_range_1.blockSignals(True)
        self.delta_spin.blockSignals(True)
        self.delta_slider.blockSignals(True)

        self.delta_spin.setValue(value)
        self.delta_slider.setValue(value)
        self.sp_range_1.setValue(value)
        self.mpl_edelta.update_delta(value)
        self.on_params_init()

        self.sp_range_1.blockSignals(False)
        self.delta_spin.blockSignals(False)
        self.delta_slider.blockSignals(False)

        # Update stored indentation depth
        if self.cb_delta_select_1.currentIndex() == 1:
            # Set indentation depth individually
            fdist = self.current_curve
            self._indentation_depth_individual[fdist] = value

    def on_export_fit_results(self):
        """Saves all fit results"""
        fname, _e = QtWidgets.QFileDialog.getSaveFileName(
            self.parent_widget,
            "Save fit results",
            "fit_results_{:03d}.tsv".format(
                self._instance_counter),
            "Tab Separated Values (*.tsv)"
        )
        if fname:
            if not fname.endswith(".tsv"):
                fname += ".tsv"
            self.on_fit_all()
            exp_curv = [fdist for fdist in self.selected_curves]
            ratings = self.rate_data(exp_curv)
            export.save_tsv_approach_retract(filename=fname,
                                             fdist_list=exp_curv,
                                             ratings=ratings)

    def on_export_edelta(self):
        """Saves all edelta curves"""
        fname, _e = QtWidgets.QFileDialog.getSaveFileName(
            self.parent_widget,
            "Save E(δ) curves",
            "",
            "Tab Separated Values (*.tsv)"
        )
        if fname:
            if not fname.endswith(".tsv"):
                fname += ".tsv"
            # Make sure all curves have correct fit properties
            self.on_fit_all()
            # Proceed computing edelta curves with progress bar
            curves = self.selected_curves
            bar = QtWidgets.QProgressDialog("Computing  E(δ) curves...",
                                            "Stop", 1, len(curves))
            bar.setWindowTitle("Please wait...")
            bar.setMinimumDuration(1000)

            res = []
            for ii, ar in enumerate(curves):
                if bar.wasCanceled():
                    return
                # TODO:
                # - Use the callback method in `compute_emodulus_mindelta`
                #   to prevent "freezing" of the GUI?
                try:
                    e, d = ar.compute_emodulus_mindelta()
                except nfit.FitDataError:
                    pass
                else:
                    res += [d, e]
                QtCore.QCoreApplication.instance().processEvents()
                bar.setValue(ii+1)

            # export curves with numpy
            with io.open(fname, "w") as fd:
                header = ["# Indentation [m] and elastic modulus [Pa]\n",
                          "# are stored as alternating rows.\n",
                          ]
                fd.writelines(header)

            with io.open(fname, "ab") as fd:
                np.savetxt(fd, np.array(res))

    def on_fit_all(self):
        """Apply initial parameters to all curves and fit"""
        # We will fit all curves with the currently visible settings
        bar = QtWidgets.QProgressDialog("Fitting all curves...",
                                        "Stop", 1, len(self.data_set))
        bar.setWindowTitle("Loading data files")
        bar.setMinimumDuration(1000)
        for ii, fdist in enumerate(self.data_set):
            QtCore.QCoreApplication.instance().processEvents()
            if bar.wasCanceled():
                break
            self.fit_apply_preprocessing(fdist)
            self.fit_approach_retract(fdist, update_ui=False)
            bar.setValue(ii+1)
            self.curve_list_update(item=ii)
        # Display map
        self.mpl_qmap_update()

    def on_rating_threshold(self):
        """(De)select curves according to threshold rating"""
        thresh = self.sp_rating_thresh.value()
        self.curve_list_update()
        for ii, ar in enumerate(self.data_set):
            rating = self.rate_data(ar)
            it = self.list_curves.topLevelItem(ii)
            if rating >= thresh:
                it.setCheckState(3, 2)
            else:
                it.setCheckState(3, 0)
        # TODO:
        # -make this more efficient. There is a lot written to disk here.
        for fdist in self.data_set:
            self.autosave(fdist)

    def on_update_weights(self, on_params_init=True):
        """Compute the weight in µm or % for user convenience

        This method updates the value of the interval range to
        suppress residuals near the contact point for the spin controls
        - `self.sp_weight_cp_um` (absolute [µm] radio button) and
        - `self.sp_weight_cp_perc` (% tip radius radio button),
        depending on which was updated by the user.

        Notes
        -----
        This method only updates the values if the model has a
        parameter "R", the radius of the indenter. In any other
        case, it will do nothing and disable the "% tip radius"
        radio button.
        """
        # First get the tip radius
        model = nmodel.get_model_by_name(self.cb_model.currentText())
        params = model.get_parameter_defaults()
        itab = self.table_parameters_initial
        for ii, key in enumerate(list(params.keys())):
            if key == "R":
                radius = float(itab.item(ii, 1).text())
                # Make sure the "% tip radius" radio button is enabled,
                # because it could have been disabled for a fit model that
                # does not contain the parameter "R".
                self.rd_weigt_perc.setEnabled(True)
                # Did the user select "% tip radius"?
                percent = self.rd_weigt_perc.isChecked()

                if percent:
                    # set um value
                    perc = self.sp_weight_cp_perc.value()
                    self.sp_weight_cp_um.blockSignals(True)
                    self.sp_weight_cp_um.setValue(radius*perc/100)
                    self.sp_weight_cp_um.blockSignals(False)
                else:
                    # set percent value
                    val = self.sp_weight_cp_um.value()
                    self.sp_weight_cp_perc.blockSignals(True)
                    self.sp_weight_cp_perc.setValue(val/radius*100)
                    self.sp_weight_cp_perc.blockSignals(False)
                break
        else:
            # The key "R" could not be found. This means we
            # have a model that does not allow to compute the
            # indentation in percentage of the tip radius
            # (e.g. conical indenter).
            # We will force the user to select the "absolute [µm]"
            # radio button and disable the "% tip radius" radio button.
            self.rd_weight_um.setChecked(True)
            self.rd_weigt_perc.setEnabled(False)

        if on_params_init:
            self.on_params_init()

    def preproc_descr_update(self, source):
        """Update the description box with the method doc string

        Writes text to `self.text_preprocessing` according to the
        method selected in `self.list_preproc_available` or
        `self.list_preproc_applied`.
        """
        if source == "available":
            thelist = self.list_preproc_available
        elif source == "applied":
            thelist = self.list_preproc_applied
        item = thelist.currentItem()
        if item is None:
            return
        method_name = item.text()
        method = getattr(npreproc.IndentationPreprocessor, method_name)
        doc = method.__doc__.replace("    ", "").strip()
        self.text_preprocessing.setPlainText(doc)

    def preproc_set_preset(self):
        """Update preselection"""
        text = self.cb_preproc_presel.currentText()
        self.list_preproc_applied.clear()
        if text == "None":
            pass
        elif text == "Recommended":
            recommended_methods = ["compute_tip_position",
                                   "correct_force_offset",
                                   "correct_tip_offset",
                                   "correct_split_approach_retract"]
            for m in recommended_methods:
                item = QtWidgets.QListWidgetItem()
                item.setText(m)
                self.list_preproc_applied.addItem(item)

    def preproc_setup(self):
        """Setup everything necessary for the preprocessing tab"""
        # Get list of preprocessing methods
        mem = inspect.getmembers(npreproc.IndentationPreprocessor)
        premem = []
        for m in mem:
            if (m[1].__doc__ is not None and
                not m[0].startswith("_") and
                    not m[0] == "apply"):
                premem.append(m[0])

        premem.sort()

        for p in premem:
            item = QtWidgets.QListWidgetItem()
            item.setText(p)
            self.list_preproc_available.addItem(item)

        self.list_preproc_available.currentItemChanged.connect(
            lambda: self.preproc_descr_update("available"))
        self.list_preproc_applied.currentItemChanged.connect(
            lambda: self.preproc_descr_update("applied"))

        # Add recommended item (see `self.preproc_set_preset`)
        self.cb_preproc_presel.addItem("Recommended")
        self.cb_preproc_presel.activated.connect(self.preproc_set_preset)
        self.cb_preproc_presel.currentIndexChanged.connect(
            self.preproc_set_preset)
        # Apply recommended defaults
        self.cb_preproc_presel.setCurrentIndex(1)

    def rate_data(self, data):
        """Apply rating to curves

        Parameters
        ----------
        data: list, afmlib.indentaion.Indentation, afmlib.AFM_DataSet
           The data to be rated
        """
        if isinstance(data, nindent.Indentation):
            data = [data]
            return_single = True
        else:
            return_single = False

        scheme_id = self.cb_rating_scheme.currentIndex()
        scheme_key = list(RATING_SCHEMES.keys())[scheme_id]
        training_set, regressor = RATING_SCHEMES[scheme_key]
        rates = []
        for fdist in data:
            rt = fdist.rate_quality(regressor=regressor,
                                    training_set=training_set)
            rates.append(rt)

        if return_single:
            return rates[0]
        else:
            return rates

    @property
    def selected_curves(self):
        """Return an IndentationGroup with all curves selected by the user"""
        curves = nanite.IndentationGroup()
        for ar in self.data_set:
            idx = self.data_set.index(ar)
            item = self.list_curves.topLevelItem(idx)
            if item.checkState(3) == 2:
                curves.append(ar)
        return curves

    def signal_slot(self, enable):
        """Enable or disable signal-slot connections"""
        cn = [
            # Signals and slots
            # Fit all button
            [self.btn_fitall.clicked, self.on_fit_all],
            # Parameter changed
            [self.cb_segment.currentTextChanged, self.on_params_init],
            [self.cb_xaxis.currentTextChanged, self.on_params_init],
            [self.cb_yaxis.currentTextChanged, self.on_params_init],
            [self.cb_model.currentTextChanged, self.on_model],
            [self.cb_range_type.currentTextChanged, self.on_params_init],
            # This does not work for checkboxes
            [self.table_parameters_initial.itemChanged, self.on_params_init],
            # Filter rating threshold
            [self.btn_rating_filter.clicked, self.on_rating_threshold],
            # Toggle weighting of contact point data
            [self.cb_weight_cp.stateChanged, self.on_params_init],
            [self.sp_weight_cp_um.valueChanged, self.on_update_weights],
            [self.sp_weight_cp_perc.valueChanged, self.on_update_weights],
            # number of samples for Edelta plot
            [self.sp_delta_num_samples.valueChanged, self.on_params_init],
            [self.sp_delta_num_samples.valueChanged, self.mpl_edelta_update],
            # rating scheme dropdown
            [self.cb_rating_scheme.currentTextChanged, self.on_params_init],
        ]

        for signal, slot in cn:
            if enable:
                signal.connect(slot)
            else:
                signal.disconnect(slot)



import pathlib
import pkg_resources
import signal
import sys
import traceback

from PyQt5 import uic, QtWidgets

from . import custom_widgets

from .. import registry
from ..settings import SettingsFile
from .._version import version as __version__

# load QMainWindow from ui file
ui_path = pkg_resources.resource_filename("pyjibe.head", "main_design.ui")
MainBase = uic.loadUiType(ui_path)[0]


class PyJibeQMdiSubWindow(QtWidgets.QMdiSubWindow):
    def closeEvent(self, QCloseEvent):
        """Correctly de-register a data set before removing the subwindow"""
        mainwidget = self.mdiArea().parentWidget().parentWidget()
        mainwidget.rem_subwindow(self.windowTitle())
        super(PyJibeQMdiSubWindow, self).closeEvent(QCloseEvent)


class PyJibe(QtWidgets.QMainWindow, MainBase):
    def __init__(self):
        QtWidgets.QMainWindow.__init__(self)
        MainBase.__init__(self)
        self.setupUi(self)
        self.setWindowTitle("PyJibe {}".format(__version__))
        # Disable native menubar (e.g. on Mac)
        self.menubar.setNativeMenuBar(False)
        # Connect menu entries
        self.action_open_bulk.triggered.connect(self.on_open_bulk)
        self.action_open_single.triggered.connect(self.on_open_single)
        self.action_open_multiple.triggered.connect(self.on_open_multiple)
        # Add settings
        self.settings = SettingsFile()

        self.subwindows = []
        self.subwindow_data = []
        self.mdiArea.cascadeSubWindows()
        self.showMaximized()

    def add_subwindow(self, widget, obj):
        """Add a subwindow, register data set and add to menu"""
        sub = PyJibeQMdiSubWindow()
        sub.setWidget(widget)
        self.mdiArea.addSubWindow(sub)
        sub.show()
        self.subwindows.append(sub)
        self.subwindow_data.append(obj)
        # Add export choices
        if hasattr(obj, "get_export_choices"):
            choices = obj.get_export_choices()
            menobj = self.menuExport.addMenu(widget.windowTitle())
            for choice in choices:
                action = menobj.addAction(choice[0])
                action.triggered.connect(getattr(obj, choice[1]))

    def rem_subwindow(self, title):
        """De-register a data set and remove from the menu"""
        for ii, sub in enumerate(self.subwindows):
            if sub.windowTitle() == title:
                self.subwindows.pop(ii)
                self.subwindow_data.pop(ii)
                break

        for action in self.menuExport.actions():
            if action.text() == title:
                self.menuExport.removeAction(action)
                break

    def on_open_bulk(self, evt=None):
        dlg = custom_widgets.FileDialog()
        search_dir = self.settings.get_path("load data")
        dlg.setDirectory(search_dir)
        if dlg.exec_():
            files = dlg.selectedFilesRecursive()
            files.sort()
            if files:
                self.load_data(files=files, retry_open=self.on_open_bulk,
                               separate_analysis=False)
                self.settings.set_path(dlg.getDirectory(), name="load data")

    def on_open_multiple(self, evt=None):
        dlg = custom_widgets.FileDialog()
        search_dir = self.settings.get_path("load data")
        dlg.setDirectory(search_dir)

        if dlg.exec_():
            files = dlg.selectedFilesRecursive()
            files.sort()
            if files:
                self.load_data(files=files, retry_open=self.on_open_multiple,
                               separate_analysis=True)
                self.settings.set_path(dlg.getDirectory(), name="load data")

    def on_open_single(self, evt=None):
        exts = ["*"+e for e in registry.file_extensions]
        exts_str = "Supported file types ({})".format(" ".join(exts))
        search_dir = self.settings.get_path("load data")

        n, _e = QtWidgets.QFileDialog.getOpenFileName(self, "Open single file",
                                                      search_dir, exts_str)
        if n:
            # user did not press cancel
            self.load_data(files=[n], retry_open=self.on_open_single)
            self.settings.set_path(pathlib.Path(n).parent, name="load data")

    def load_data(self, files, retry_open=None, separate_analysis=False):
        # approach-retract data files
        supfiles = []
        for ff in files:
            path = pathlib.Path(ff)
            if path.suffix in registry.file_extensions:
                supfiles.append(ff)

        # check if AFM files were found
        if not len(supfiles):
            msg = QtWidgets.QMessageBox()
            msg.setIcon(QtWidgets.QMessageBox.Warning)
            msg.setText("No AFM data files found!")
            msg.setInformativeText("")
            msg.setWindowTitle("No AFM data found!")
            msg.setStandardButtons(
                QtWidgets.QMessageBox.Ok | QtWidgets.QMessageBox.Retry)
            retval = msg.exec_()
            if (retry_open is not None and
                    retval == QtWidgets.QMessageBox.Retry):
                retry_open()
        else:
            if separate_analysis:
                # open each file in one analysis
                supfiles = [[ss] for ss in supfiles]
            else:
                supfiles = [supfiles]
            for flist in supfiles:
                widget = QtWidgets.QWidget()
                fdist = registry.fd.UiForceDistance(widget)
                fdist.add_files(flist)
                self.add_subwindow(widget, fdist)
                # Add to self, otherwise events will not be triggered!
                setattr(self, "subwindow_{}".format(
                    len(self.subwindows)), fdist)


def excepthook(etype, value, trace):
    """
    Handler for all unhandled exceptions.

    Parameters
    ----------
    etype: Exception
        exception type (`SyntaxError`, `ZeroDivisionError`, etc...)
    value: str
        exception error message
    trace:
        traceback header, if any (otherwise, it prints the
        standard Python header: ``Traceback (most recent call last)``
    """
    vinfo = "Unhandled exception in PyJibe version {}:\n".format(__version__)
    tmp = traceback.format_exception(etype, value, trace)
    exception = "".join([vinfo]+tmp)

    errorbox = QtWidgets.QMessageBox()
    errorbox.addButton(QtWidgets.QPushButton('Close'),
                       QtWidgets.QMessageBox.YesRole)
    errorbox.addButton(QtWidgets.QPushButton(
        'Copy text && Close'), QtWidgets.QMessageBox.NoRole)
    errorbox.setText(exception)
    ret = errorbox.exec_()
    if ret == 1:
        cb = QtWidgets.QApplication.clipboard()
        cb.clear(mode=cb.Clipboard)
        cb.setText(exception)


# Make Ctr+C close the app
signal.signal(signal.SIGINT, signal.SIG_DFL)
# Display exception hook in separate dialog instead of crashing
sys.excepthook = excepthook



from matplotlib.figure import Figure
import matplotlib.gridspec as gridspec
from matplotlib.backends.backend_qt5agg import (
    FigureCanvasQTAgg as FigureCanvas)
import numpy as np


from .. import units
from ..head import custom_widgets


class MPLIndentation():
    def __init__(self):
        """Matplotlib plot for force-indentation data"""
        # Add matplotlib figure
        self.figure = Figure(facecolor="none", tight_layout=True,
                             frameon={"pad": .0})
        gs = gridspec.GridSpec(2, 1, height_ratios=[4, 1])

        # main axis
        self.axis_main = self.figure.add_subplot(gs[0])
        self.axis_main.grid()
        self.plots = {}
        self.plots["fit range"] = self.axis_main.axvspan(np.pi,
                                                         np.sqrt(2),
                                                         facecolor="#FDFDF0",
                                                         edgecolor="none",
                                                         label="fit range")
        self.plots["approach"] = self.axis_main.plot(range(10),
                                                     range(10),
                                                     color="#B1BDEF",
                                                     lw=2,
                                                     label="approach")[0]
        self.plots["retract"] = self.axis_main.plot(range(10),
                                                    range(10),
                                                    color="#EFB3B1",
                                                    lw=2,
                                                    label="retract")[0]
        self.plots["fit"] = self.axis_main.plot(range(10),
                                                range(10),
                                                color="blue",
                                                label="fit range")[0]

        # residuals
        self.axis_res = self.figure.add_subplot(gs[1], sharex=self.axis_main)
        self.axis_res.grid()
        self.plots["residuals"] = self.axis_res.plot(range(2),
                                                     range(2),
                                                     label="residuals")[0]

        self.canvas = FigureCanvas(self.figure)

        self.canvas.draw()

    def add_toolbar(self, widget):
        """Add toolbar to PyQT widget"""
        self.toolbar = custom_widgets.NavigationToolbarIndent(
            self.canvas,
            widget,
            coordinates=True
        )

        self.toolbar.save_data_callback = self.save_data_callback

    def save_data_callback(self, filename):
        self.fdist.export(filename)

    def update(self, fdist, rescale_x=None, rescale_y=None):
        self.fdist = fdist
        xaxis = "tip position"
        yaxis = "force"
        xscale = units.hrscale(xaxis)
        yscale = units.hrscale(yaxis)
        xunit = units.hrunit(xaxis)
        yunit = units.hrunit(yaxis)

        # approach and retract data
        for segment in ["approach", "retract"]:
            segment_bool = segment == "retract"
            self.plots[segment].set_data(
                fdist[xaxis][fdist["segment"] == segment_bool]*xscale,
                fdist[yaxis][fdist["segment"] == segment_bool]*yscale)

        self.axis_res.set_xlabel("{} [{}]".format(xaxis, xunit))
        self.axis_res.set_ylabel("residuals [{}]".format(yunit))
        self.axis_main.set_ylabel("{} [{}]".format(yaxis, yunit))

        if "fit" in fdist.data:
            self.plots["residuals"].set_visible(True)
            self.plots["fit"].set_visible(True)

            self.plots["fit"].set_data(fdist["tip position"]*xscale,
                                       fdist["fit"]*yscale)
            self.plots["residuals"].set_data(fdist["tip position"]*xscale,
                                             (fdist["fit residuals"])*yscale)
            # fit range
            xy = self.plots["fit range"].get_xy()
            fitrange = (fdist[xaxis]*xscale)[fdist["fit range"]]
            fitmin = np.min(fitrange)
            fitmax = np.max(fitrange)
            xy[:, 0] = fitmax
            xy[2:4, 0] = fitmin
            self.plots["fit range"].set_xy(xy)
        else:
            self.plots["residuals"].set_visible(False)
            self.plots["fit"].set_visible(False)

        self.update_plot(rescale_x=rescale_x,
                         rescale_y=rescale_y)

    def update_plot(self, rescale_x=None, rescale_y=None):
        """Update plot data range"""
        if rescale_x is None:
            fit_range = self.fdist["fit range"].values
            xmin = np.min(self.plots["fit"].get_data()[0][fit_range])
            xmax = np.max(self.plots["fit"].get_data()[0][fit_range])
            xmargin = np.abs(xmax - xmin) * .05
            xmin -= xmargin
            xmax += xmargin
        else:
            xmin, xmax = rescale_x

        if xmin == xmax:
            xmin = xmax = np.nan

        if rescale_y is None:
            fit_range = self.fdist["fit range"].values
            ymin = np.min(self.plots["fit"].get_data()[1][fit_range])
            ymax = np.max(self.plots["fit"].get_data()[1][fit_range])
            ymargin = np.abs(ymax - ymin) * .05
            ymin -= ymargin
            ymax += ymargin
        else:
            ymin, ymax = rescale_y

        if ymin == ymax:
            ymin = ymax = np.nan

        if not np.isnan(xmin + xmax):
            # x: main plot and residuals
            axes = self.axis_main, self.axis_res
            for ax in axes:
                ax.set_xlim(xmin, xmax)
        if not np.isnan(ymin + ymax):
            # y: main plot
            self.axis_main.set_ylim(ymin, ymax)
        # set residuals ylim automatically
        rmin = np.min(self.plots["residuals"].get_data()[1])
        rmax = np.max(self.plots["residuals"].get_data()[1])
        if not np.isnan(rmin + rmax):
            rmax = max(abs(rmin), abs(rmax))
            rmax = np.ceil(rmax*2) / 2
            rmin = -rmax
            # y: main plot
            self.axis_res.set_ylim(rmin, rmax)

        self.canvas.draw()



import numpy as np


def equals(a, b):
    """Check if `a` equals `b` with support for arrays"""
    if a.__class__ != b.__class__:
        return False
    if isinstance(a, dict):
        if len(a) != len(b):
            return False
        ret = True
        for key in a:
            if key not in b:
                return False
            else:
                ret *= equals(a[key], b[key])
        return ret
    elif isinstance(a, (list, tuple)):
        if len(a) != len(b):
            return False
        ret = True
        for ii in range(len(a)):
            ret *= equals(a[ii], b[ii])
        return ret
    elif isinstance(a, np.ndarray):
        if len(a) != len(b):
            return False
        else:
            return np.all(a == b)
    else:
        return a == b



import io

from matplotlib.figure import Figure
from matplotlib.backends.backend_qt5agg import (
    FigureCanvasQTAgg as FigureCanvas)

import nanite.fit as nfit
import numpy as np
from PyQt5.QtWidgets import QApplication


from .. import units
from ..head import custom_widgets


class MPLEDelta(object):
    def __init__(self):
        self.figure = Figure(facecolor="none", tight_layout=False)
        # verbose text if no map data is available
        self.no_data_text = self.figure.text(.5, .5,
                                             "no E(δ) data available",
                                             horizontalalignment="center")
        self.no_data_text.set_visible(False)
        # main axis
        self.axis_main = self.figure.add_subplot(111)
        self.axis_main.grid()
        self.plot = self.axis_main.plot(range(10), range(10))[0]
        self.opt_auto = self.axis_main.axvline(5, c="r")
        self.canvas = FigureCanvas(self.figure)

        self.canvas.draw()
        se = units.hrunit("young's modulus")
        sd = units.hrunit("tip radius")
        self.axis_main.set_xlabel("indentation depth δ [{}]".format(sd))
        self.axis_main.set_ylabel("elastic modulus E [{}]".format(se))
        # adjust suplot parameters left,bottom,right,top
        self.figure.subplots_adjust(0.2, 0.15, 0.95, 0.95)

        # Update parameters
        self._update_in_progress_locks = {}
        self._update_in_progress_active = None

    def add_toolbar(self, widget):
        self.toolbar = custom_widgets.NavigationToolbarEDelta(
            self.canvas,
            widget,
            coordinates=True
        )
        self.toolbar.save_data_callback = self.save_data_callback

    def hide_plot(self, boolval):
        self.axis_main.set_visible(not boolval)
        self.no_data_text.set_visible(boolval)
        self.toolbar.setDisabled(boolval)

    def save_data_callback(self, filename):
        """Save current image as tsv"""
        with io.open(filename, "w") as fd:
            fd.write("indentation depth [m]\telastic modulus [Pa]\n")
        with io.open(filename, "ab") as fd:
            np.savetxt(fd, self.plot_data, delimiter="\t")

    def update(self, fdist, delta_opt=None):
        """Update the map tab plot data

        Parameters
        ----------
        fdist: afmlib.Indentation
            Approach-Retract data set
        delta_opt: float
            Optimal indentation depth
        """
        # TODO:
        # - use Python threading instead of this lambda method?
        self._update_in_progress_active = fdist
        if fdist in self._update_in_progress_locks:
            QApplication.instance().processEvents()
        else:
            self._update_in_progress_locks[fdist] = True
            def cbfdist(e, d): return self.update_plot(e, d, fdist=fdist)
            try:
                emod, delta = fdist.compute_emodulus_mindelta(callback=cbfdist)
            except nfit.FitDataError:
                # Could not generate E(d) plot due to weird data
                self.hide_plot(True)
                self.canvas.draw()
            else:
                self.update_plot(emod, delta, delta_opt, fdist=fdist)
            self._update_in_progress_locks.pop(fdist)

    def update_delta(self, delta):
        """Updates the vertical line for indentation depth"""
        self.opt_auto.set_data(delta, (0, 1))
        self.canvas.draw()

    def update_plot(self, emoduli, indentations,
                    delta_opt=None, fdist=None):
        """
        Parameters
        ----------
        emoduli: 1d ndarray
            emodulus values
        mindelta: 1d ndarray
            minimal indentation
        """
        self.hide_plot(False)
        if self._update_in_progress_active is fdist:
            fact_emod = units.hrscale("young's modulus")
            fact_delt = units.hrscale("tip radius")

            emoduli_p = emoduli*fact_emod
            mindeltas_p = indentations*fact_delt
            self.plot.set_data(mindeltas_p, emoduli_p)
            self.axis_main.set_xlim(mindeltas_p.min(), mindeltas_p.max())
            self.axis_main.set_ylim(emoduli_p.min(), emoduli_p.max())
            if delta_opt is not None:
                self.opt_auto.set_data(delta_opt, (0, 1))
            self.canvas.draw()

            self.plot_data = np.zeros((emoduli.size, 2), dtype=float)
            self.plot_data[:, 0] = indentations
            self.plot_data[:, 1] = emoduli

            QApplication.processEvents()
